@@ -24,6 +24,7 @@
 
 #include <string.h>
 #include <stdio.h>
+#include "gc.h"
 #include "gobject.h"
 #include "private.h"
 #include "guile-support.h"
@@ -51,6 +52,14 @@ SCM_SYMBOL  (sym_gruntime_error,"gruntime-error");
 
 #define DEBUG_REFCOUNTING
 
+static const scm_t_gtype_instance_funcs gobject_funcs = {
+    G_TYPE_OBJECT,
+    (scm_t_gtype_instance_ref)g_object_ref,
+    (scm_t_gtype_instance_unref)g_object_unref,
+    (scm_t_gtype_instance_get_qdata)g_object_get_qdata,
+    (scm_t_gtype_instance_set_qdata)g_object_set_qdata
+};
+
 
 
 SCM_DEFINE (scm_gobject_set_data_x, "gobject-set-data!", 3, 0, 0,
@@ -71,8 +80,8 @@ SCM_DEFINE (scm_gobject_set_data_x, "gobject-set-data!", 3, 0, 0,
     else
         g_object_set_qdata_full
             (gobject, g_quark_from_string (sym),
-             GINT_TO_POINTER (SCM_UNPACK (scm_gc_protect_object (val))),
-             (GDestroyNotify)scm_gc_unprotect_object);
+             SCM_TO_GPOINTER (scm_glib_gc_protect_object (val)),
+             (GDestroyNotify)scm_glib_gc_unprotect_object);
         
     return SCM_UNSPECIFIED;
 }
@@ -95,7 +104,7 @@ SCM_DEFINE (scm_gobject_get_data, "gobject-get-data", 2, 0, 0,
     data = g_object_get_qdata (gobject, g_quark_from_string (sym));
     
     if (data)
-        return SCM_PACK (GPOINTER_TO_INT (data));
+        return GPOINTER_TO_SCM (data);
     else
         return SCM_UNBOUND;
 }
@@ -185,7 +194,7 @@ scm_c_gtype_instance_class_init (gpointer g_class, gpointer class_data)
     DEBUG_ALLOC ("  protecting class %p of %s gclass %p", class,
                  g_type_name (G_TYPE_FROM_CLASS (g_class)), class);
 
-    guile_class->class = scm_gc_protect_object (class);
+    guile_class->class = scm_glib_gc_protect_object (class);
 
     /* Not calling a class-init generic will prevent GOOPS classes that are
      * subclassed on the scheme side from being initialized, but that's a corner
@@ -358,7 +367,7 @@ SCM_DEFINE (scm_gobject_class_install_property, "gobject-class-install-property"
     DEBUG_ALLOC ("  protecting param %p of %s gparam %p", param,
                  g_type_name (G_TYPE_FROM_INSTANCE (gparam)), gparam);
     g_hash_table_insert (guile_class->properties_hash, GINT_TO_POINTER (id),
-			 scm_gc_protect_object (param));
+			 scm_glib_gc_protect_object (param));
 
     scm_call_1 (_gobject_class_set_properties_x, class);
 
@@ -534,7 +543,7 @@ SCM_DEFINE (scm_gobject_primitive_create_instance, "gobject-primitive-create-ins
     scm_slot_set_x (object, scm_sym_gtype_instance, smob);
     
     /* cache this wrapper, like in scm_c_gtype_instance_to_scm */
-    g_object_set_qdata (gobject, guile_gobject_quark_instance_wrapper, object);
+    scm_c_gtype_instance_set_cached_goops (gobject, object);
 
     return SCM_UNSPECIFIED;
 }
@@ -576,9 +585,10 @@ scm_init_gnome_gobject (void)
 #ifndef SCM_MAGIC_SNARFER
 #include "gobject.x"
 #endif
-    _gobject_initargs_fluid = scm_make_fluid ();
+    scm_register_gtype_instance_funcs (&gobject_funcs);
 
-    /* there is a case where it won't be set before entering
+    _gobject_initargs_fluid = scm_make_fluid ();
+    /* there is a case where the fluid won't be set before entering
        scm_c_gtype_instance_instance_init: if the class is instantiated from C
        via g_object_new instead of from scheme via `make'. Give the initargs a
        sane value in that case. */
