@@ -26,6 +26,7 @@
 
 #include <stdio.h>
 #include <string.h>
+#include "gc.h"
 #include "gparameter.h"
 #include "guile-support.h"
 
@@ -35,6 +36,14 @@ SCM scm_class_gparam;
 
 static SCM scm_gparam_struct_vtable;
 
+static const scm_t_gtype_instance_funcs gparamspec_funcs = {
+    G_TYPE_PARAM,
+    (scm_t_gtype_instance_ref)g_param_spec_ref,
+    (scm_t_gtype_instance_unref)g_param_spec_unref,
+    (scm_t_gtype_instance_get_qdata)g_param_spec_get_qdata,
+    (scm_t_gtype_instance_set_qdata)g_param_spec_set_qdata
+};
+
 
 
 #define scm_gparam_struct_i_name		(scm_vtable_offset_user)
@@ -323,8 +332,8 @@ SCM_DEFINE (scm_gparam_to_param_struct, "gparam->param-struct", 1, 0, 0,
                     " be unreffed"); */
 
     g_param_spec_set_qdata_full (pspec, quark_param_struct,
-				 scm_gc_protect_object (pspec_struct),
-				 (GDestroyNotify) scm_gc_unprotect_object);
+				 scm_glib_gc_protect_object (pspec_struct),
+				 (GDestroyNotify) scm_glib_gc_unprotect_object);
 
     return pspec_struct;
 }
@@ -628,6 +637,8 @@ scm_init_gnome_gobject_parameters (void)
 #include "gparameter.x"
 #endif
 
+    scm_register_gtype_instance_funcs (&gparamspec_funcs);
+
     quark_param_struct = g_quark_from_static_string ("%scm-param-struct");
 
     gsubr = scm_c_make_gsubr ("%print-gparam-struct", 2, 0, 0, print_gparam_struct);
