@@ -26,7 +26,9 @@
 
 #include <stdio.h>
 #include <string.h>
+#include "gc.h"
 #include "gvalue.h"
+#include "gobject.h"
 #include "guile-support.h"
 
 
@@ -55,7 +57,7 @@ static gpointer
 copy_gboxed_scm (gpointer boxed)
 {
     DEBUG_ALLOC (G_STRLOC ": copying gboxed %p", boxed);
-    scm_gc_protect_object ((SCM) boxed);
+    scm_glib_gc_protect_object ((SCM) boxed);
     return boxed;
 }
 
@@ -63,7 +65,7 @@ static void
 free_gboxed_scm (gpointer boxed)
 {
     DEBUG_ALLOC (G_STRLOC ": freeing gboxed %p", boxed);
-    scm_gc_unprotect_object ((SCM) boxed);
+    scm_glib_gc_unprotect_object ((SCM) boxed);
 }
 
 GType
@@ -256,7 +258,7 @@ SCM_DEFINE (scm_gvalue_primitive_set, "gvalue-primitive-set", 2, 0, 0,
 	if (SCM_FALSEP (value))
 	    g_value_set_string (gvalue, NULL);
 	else
-	    g_value_set_string (gvalue, g_strdup (SCM_STRING_CHARS (value)));
+	    g_value_set_string (gvalue, SCM_STRING_CHARS (value));
 	break;
 
     case G_TYPE_ENUM: {
@@ -292,17 +294,22 @@ SCM_DEFINE (scm_gvalue_primitive_set, "gvalue-primitive-set", 2, 0, 0,
         if (SCM_FALSEP (value)) {
             g_value_set_object (gvalue, NULL);
         } else {
-            GTypeInstance *ginstance;
+            gpointer ginstance;
             GType gtype;
 
-            SCM_VALIDATE_GTYPE_INSTANCE_COPY (2, value, ginstance);
+            if (SCM_GOBJECTP (value)) {
+                SCM_VALIDATE_GOBJECT_COPY (2, value, ginstance);
+            }
+            else {
+                SCM_VALIDATE_GTYPE_INSTANCE_COPY (2, value, ginstance);
+            }
             gtype = G_TYPE_FROM_INSTANCE (ginstance);
 
             switch (G_TYPE_FUNDAMENTAL (gtype)) {
             case G_TYPE_OBJECT:
                 SCM_ASSERT (g_type_is_a (gtype, G_VALUE_TYPE (gvalue)),
                             value, SCM_ARG2, FUNC_NAME);
-                g_value_set_object (gvalue, G_OBJECT (ginstance));
+                g_value_set_object (gvalue, ginstance);
                 break;
             default:
                 SCM_ERROR_NOT_YET_IMPLEMENTED (value);
@@ -314,8 +321,7 @@ SCM_DEFINE (scm_gvalue_primitive_set, "gvalue-primitive-set", 2, 0, 0,
     case G_TYPE_BOXED:
         if (G_VALUE_TYPE (gvalue) == G_TYPE_BOXED_SCM) {
             /* the copy func will protect the scm_value from GC */
-            g_value_set_boxed (gvalue,
-                               GINT_TO_POINTER (SCM_UNPACK (value)));
+            g_value_set_boxed (gvalue, SCM_TO_GPOINTER (value));
         } else if (G_VALUE_TYPE (gvalue) == G_TYPE_VALUE_ARRAY) {
             GValueArray *arr;
             gint len;
@@ -327,10 +333,10 @@ SCM_DEFINE (scm_gvalue_primitive_set, "gvalue-primitive-set", 2, 0, 0,
             arr = g_value_array_new (len);
             while (len--) {
                 GType value_type;
-                SCM val, v;
+                SCM v;
 
                 v = SCM_CAR (value);
-
+                
                 if (SCM_STRINGP (v))
                     value_type = G_TYPE_STRING;
                 else if (SCM_BOOLP (v))
@@ -341,15 +347,18 @@ SCM_DEFINE (scm_gvalue_primitive_set, "gvalue-primitive-set", 2, 0, 0,
                     value_type = G_TYPE_DOUBLE;
                 else if (SCM_CHARP (v))
                     value_type = G_TYPE_CHAR;
-                else if (scm_list_p (v))
+                else if (SCM_GOBJECTP (v)) {
+                    GObject *gobject;
+                    SCM_VALIDATE_GOBJECT_COPY (1, v, gobject);
+                    value_type = G_OBJECT_TYPE (gobject);
+                }
+                else if (SCM_BOOL (scm_list_p (v)))
                     value_type = G_TYPE_VALUE_ARRAY;
                 else
                     scm_wrong_type_arg (FUNC_NAME, SCM_ARG1, v);
 
-                val = scm_c_make_gvalue (value_type);
-                scm_gvalue_primitive_set (val, v);
                 /* will copy the val */
-                g_value_array_append (arr, (GValue*) SCM_SMOB_DATA (val));
+                g_value_array_append (arr, scm_c_scm_to_gvalue (value_type, v));
                 value = SCM_CDR (value);
             }
                 
@@ -428,7 +437,7 @@ SCM_DEFINE (scm_gvalue_primitive_get, "gvalue-primitive-get", 1, 0, 0,
         if (G_VALUE_TYPE (gvalue) == G_TYPE_BOXED_SCM) {
             if (!p)
                 return SCM_UNSPECIFIED;
-            return SCM_PACK (GPOINTER_TO_INT (p));
+            return GPOINTER_TO_SCM (p);
         } else if (G_VALUE_TYPE (gvalue) == G_TYPE_VALUE_ARRAY) {
             GValueArray *arr = p;
             gint i = arr ? arr->n_values : 0;
@@ -450,8 +459,8 @@ SCM_DEFINE (scm_gvalue_primitive_get, "gvalue-primitive-get", 1, 0, 0,
     }
 
     case G_TYPE_OBJECT:
-        /* scm_c_make_gtype_instance will ref the object for us */
-	return scm_c_make_gtype_instance (g_value_get_object (gvalue));
+        /* scm_c_gtype_instance_to_scm will ref the object for us */
+	return scm_c_gtype_instance_to_scm (g_value_get_object (gvalue));
 
     case G_TYPE_PARAM:
         /* scm_c_make_gtype_instance will ref the object for us */
@@ -636,6 +645,7 @@ SCM_DEFINE (scm_scm_to_gvalue, "scm->gvalue", 2, 0, 0,
                 return svalue;
             } else {
                 g_free (v);
+                scm_wrong_type_arg (FUNC_NAME, 2, scm);
             }
         }
         /* might fall through */
@@ -876,8 +886,6 @@ scm_init_gnome_gobject_values (void)
     scm_c_define_and_export_gtype_x (G_TYPE_ULONG);
     scm_c_define_and_export_gtype_x (G_TYPE_INT64);
     scm_c_define_and_export_gtype_x (G_TYPE_UINT64);
-    scm_c_define_and_export_gtype_x (G_TYPE_ENUM);
-    scm_c_define_and_export_gtype_x (G_TYPE_FLAGS);
     scm_c_define_and_export_gtype_x (G_TYPE_FLOAT);
     scm_c_define_and_export_gtype_x (G_TYPE_DOUBLE);
     scm_c_define_and_export_gtype_x (G_TYPE_STRING);
