@@ -28,18 +28,17 @@
 (define-module (gnome gw support gobject)
   #:use-module (oop goops)
   #:use-module (ice-9 optargs)
-  #:use-module (ice-9 slib)
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-11)
   #:use-module (srfi srfi-13)
   
-  #:use-module (g-wrap)
+  #:use-module (gnome gw support g-wrap)
   #:use-module (g-wrap enumeration)
   #:use-module (g-wrap rti)
-  #:use-module (g-wrap guile)
   #:use-module (g-wrap c-types)
   
   #:use-module (gnome gobject utils)
+  #:use-module (gnome gw support slib)
   
   #:export (unwrap-null-checked
             
@@ -64,14 +63,19 @@
             wrap-flags!
             wrap-gobject-class!
             
-            wrap-custom-boxed!))
-
-(require 'printf)
+            wrap-custom-boxed!
+            wrap-custom-gvalue!))
 
 (define-class <gobject-wrapset-base> (<gw-guile-wrapset>)
   (type-aliases #:init-form (make-hash-table 31))
   (type-rules #:init-form (make-hash-table 7)))
 
+(define-method (initializations-cg (wrapset <gobject-wrapset-base>) err)
+  (list
+   (next-method)
+   "gw_guile_set_generics_module_x (scm_c_resolve_module\n"
+   "                                (\"gnome gw generics\"));\n"))
+
 (define-method (add-type-alias! (wrapset <gobject-wrapset-base>)
                                 (alias <string>)
                                 (name <symbol>))
@@ -179,7 +183,8 @@
 (define-method (add-type! (ws <gobject-wrapset-base>)
                           (type <gobject-classed-type>))
   (next-method)
-  (add-module-export! ws (class-name type)))
+  (if (slot-ref type 'define-class?)
+      (add-module-export! ws (class-name type))))
 
 (define-class <gobject-classed-pointer-type> (<gobject-classed-type>))
 
@@ -242,7 +247,7 @@
      (if-typespec-option value 'caller-owned
          ;; the _to_scm will ref the object; if the function is a
          ;; constructor, we don't need that ref
-          (list "if (" c-var ") g_object_unref ((GObject*)" c-var ");\n")))))
+          (list "if (" c-var ") scm_c_gtype_instance_unref ((GTypeInstance*)" c-var ");\n")))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;; Wrap boxed types, represented on the scheme side by GValues.
@@ -299,7 +304,12 @@
      "  " scm-var " = SCM_BOOL_F;\n"
      "} else {\n"
      "  " scm-var " = scm_c_make_gvalue (" (gtype-id type) ");\n"
-     "  g_value_set_boxed ((GValue *) SCM_SMOB_DATA (" scm-var "), " c-var ");\n"
+     (if-typespec-option
+      value 'callee-owned
+      (list
+       "  g_value_set_boxed ((GValue *) SCM_SMOB_DATA (" scm-var "), " c-var ");\n")
+      (list
+       "  g_value_take_boxed ((GValue *) SCM_SMOB_DATA (" scm-var "), " c-var ");\n"))
      "}\n")))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -675,3 +685,55 @@
        (slot-set! t 'unwrap (,make-custom-unwrapper t ,unwrap))
        (add-type! ws t)
        (add-type-alias! ws ,pname (name t)))))
+
+
+(define-class <gobject-custom-gvalue-type> (<gobject-classed-type>)
+  (wrap-func #:init-keyword #:wrap-func #:getter wrap-func)
+  (unwrap-func #:init-keyword #:unwrap-func #:getter unwrap-func)
+  
+  #:allowed-options '(null-ok))
+
+(define-method (make-typespec (type <gobject-custom-gvalue-type>) (options <list>))
+  (next-method type (cons 'unspecialized options)))
+
+(define-method (initializations-cg (wrapset <gobject-wrapset-base>)
+                                   (type <gobject-custom-gvalue-type>)
+                                   status-var)
+  (list
+   (next-method)
+   "scm_c_register_gvalue_wrappers (" (gtype-id type) ", "
+   (wrap-func type) ", " (unwrap-func type) ");\n"))
+
+(define-method (unwrap-value-cg (type <gobject-custom-gvalue-type>)
+                                (value <gw-value>)
+                                status-var)
+  (let ((c-var (var value))
+        (scm-var (scm-var value)))
+    (unwrap-null-checked
+     value status-var
+     (list
+      c-var " = g_new0 (GValue, 1);\n"
+      (unwrap-func type) " (" scm-var ", " c-var ");\n"
+      "if (!G_IS_VALUE (" c-var ")) {"
+      "  g_free (" c-var ");\n"
+      "  " c-var " = NULL;\n"
+      "}\n"))))
+
+(define-method (wrap-value-cg (type <gobject-custom-gvalue-type>)
+                              (value <gw-value>)
+                              status-var)
+  (let ((c-var (var value))
+        (scm-var (scm-var value)))
+    (list
+     scm-var " = " (wrap-func type) " (" c-var ");\n")))
+
+(define-macro (wrap-custom-gvalue! ctype gtype wrap-func unwrap-func)
+  `(let ((t (make ,<gobject-custom-gvalue-type>
+              #:ctype ,ctype
+              #:gtype-id ,gtype
+              #:c-type-name "GValue*"
+              #:ffspec 'pointer
+              #:wrapped "Custom"
+              #:wrap-func ,wrap-func
+              #:unwrap-func ,unwrap-func)))
+     (add-type! ws t)))
